@@ -4,16 +4,18 @@ from pathlib import Path
 from subprocess import Popen
 from argparse import ArgumentParser, RawTextHelpFormatter
 import concurrent.futures
-import glob2
+import re
 import time
 import json
 import os
-import re
+import glob2
 
 
 def parse_arguments():
     parser = ArgumentParser(
-        description="Utility to manipulate compilation databases.",
+        description="""
+        Utility to manipulate compilation databases. (CDB)
+        https://github.com/qdewaghe/compile-commands""",
         usage="compile-commands --file=FILE",
         formatter_class=RawTextHelpFormatter,
     )
@@ -27,14 +29,17 @@ def parse_arguments():
     parser.add_argument(
         "--file",
         type=str,
-        help="path to compilation database"
+        help="path to compilation database",
     )
 
     parser.add_argument(
         "-m",
         "--merge",
         action="store_true",
-        help="find all compile-commands.json files in --dir recursively, if not only the root directory will be considered",
+        help=(
+            "find all compile-commands.json files in --dir recursively and merges them,"
+            "\nif not set only the CDB in the root directory will be considered"
+        ),
     )
 
     parser.add_argument(
@@ -61,7 +66,21 @@ def parse_arguments():
         "--remove_files",
         default="",
         type=str,
-        help="comma-seperated list of files to be removed",
+        help="comma-separated list of files to be removed from the CDB",
+    )
+
+    parser.add_argument(
+        "--include_files",
+        default="",
+        type=str,
+        help="comma-separated list of files to keep in the CDB (every other files will be removed)",
+    )
+
+    parser.add_argument(
+        "--path_prefix",
+        default="",
+        type=str,
+        help="file path prefix of include_files and remove_files",
     )
 
     parser.add_argument(
@@ -79,8 +98,8 @@ def parse_arguments():
     parser.add_argument(
         "--replacement",
         type=str,
-        default='',
-        help='replacement for matches of --filter, can reference groups matched.'
+        default="",
+        help="replacement for matches of --filter, can reference groups matched.",
     )
 
     parser.add_argument(
@@ -123,19 +142,28 @@ def parse_arguments():
         "--allow_duplicates",
         default=False,
         action="store_true",
-        help="by default duplicated files are deleted."
+        help="by default duplicated files are deleted.",
     )
 
     parser.add_argument(
-        "--version",
-        action="store_true",
-        help="prints the version"
+        "--force_write",
+        default=False,
+        help="force write compile commands even when compile commands list is empty",
     )
 
+    parser.add_argument(
+        "--absolute_include_paths",
+        default=False,
+        action="store_true",
+        help="If the include paths inside the commands are relative, make them absolute.",
+    )
+
+    parser.add_argument("--version", action="store_true", help="prints the version")
+
     args = parser.parse_args()
 
     if args.version:
-        print('compile-commands: 1.1.3')
+        print("compile-commands: 1.1.3")
         exit(0)
 
     if not args.dir and not args.file:
@@ -163,11 +191,10 @@ def parse_arguments():
 def dir_path(path):
     if os.path.isdir(path):
         return path
-    else:
-        raise NotADirectoryError(path)
+    raise NotADirectoryError(path)
 
 
-def get_compile_dbs(dir):
+def get_compile_dbs(directory):
     # TODO:
     # Replace this call by something that stop the recursion for a given directory
     # once the file has been found (BFS)
@@ -181,8 +208,9 @@ def get_compile_dbs(dir):
     #    │   └── some_directory //visit because it could be a build directory
     #    |       └── another_directory //do not visit
     #    ...
-    paths = list(glob2.glob(
-        '{}/**/compile_commands.json'.format(dir), recursive=True))
+    paths = list(
+        glob2.glob("{}/**/compile_commands.json".format(directory), recursive=True)
+    )
 
     # Since we take into account symlinks we have to make sure the symlinks
     # doesn't resolve to a file that we already take into account
@@ -192,14 +220,17 @@ def get_compile_dbs(dir):
     # if the file in the root directory is a symlink to a build directory
     # it will still be taken into account as long as the build directory
     # is inside the tree
-    return [p for p in paths if Path(
-        p).parent.parts[-1] != Path(dir).parts[-1]]
+    return [p for p in paths if Path(p).parent.parts[-1] != Path(directory).parts[-1]]
 
 
 def remove_files(data, files):
     return [d for d in data if d["file"] not in files]
 
 
+def include_files(data, files):
+    return [d for d in data if d["file"] in files]
+
+
 def merge_json_files(paths):
     data = []
     for path in paths:
@@ -230,16 +261,16 @@ def change_compiler_path(data, new_path: str):
 def to_clang(data):
     for entry in data:
         entry["command"] = (
-            entry["command"].replace(
-                "/gcc", "/clang").replace("/g++", "/clang++")
+            entry["command"].replace("/gcc", "/clang").replace("/g++", "/clang++")
         )
     return data
 
 
 def to_gcc(data):
     for entry in data:
-        entry["command"] = (entry["command"].replace(
-            "/clang++", "/g++").replace("/clang", "/gcc"))
+        entry["command"] = (
+            entry["command"].replace("/clang++", "/g++").replace("/clang", "/gcc")
+        )
     return data
 
 
@@ -257,6 +288,32 @@ def execute(data, threads, quiet):
             executor.submit(run, entry["command"], index, total, quiet)
 
 
+def absolute_include_paths(data):
+    for entry in data:
+        directory = entry["directory"]
+        command = entry["command"]
+
+        # Include paths with spaces `-I include` and without spaces `-Iinclude`
+        # are treated separately. The first regular expressions deals with the includes
+        # with spaces, and conversely for the second.
+        #
+        # This won't work well if there are multiple spaces between the include flag
+        # and the first char because it'll break the lookahead that checks that the path
+        # is indeed relative by verifying that the first char is not a '/'.
+        # There is probably a better way to do this.
+        command = re.sub(
+            r"(-I|-iquote|-isystem|-idirafter)(?=\s)(\s+)(?=[^\/])([^\/]\S*)",
+            f"\\1\\2{directory}/\\3",
+            command,
+        )
+        entry["command"] = re.sub(
+            r"(-I|-iquote|-isystem|-idirafter)(?=\S)(?=[^\/])([^\/]\S*)",
+            f"\\1{directory}/\\2",
+            command,
+        )
+    return data
+
+
 def remove_trailing(string: str, to_remove: str):
     if string.endswith(to_remove):
         return string[: -len(to_remove)]
@@ -270,7 +327,8 @@ def filter_files(data, regex: str):
 def filter_commands(data, regex, replacement):
     for entry in data:
         entry["command"] = re.sub(
-            regex, replacement, entry["command"], flags=re.IGNORECASE)
+            regex, replacement, entry["command"], flags=re.IGNORECASE
+        )
     return data
 
 
@@ -289,23 +347,28 @@ def main():
         data = merge_json_files(get_compile_dbs(args.dir))
     else:
         # if --merge is not set we use existing data inside the specified directory
-        filepath = '{}/compile_commands.json'.format(args.dir)
+        filepath = "{}/compile_commands.json".format(args.dir)
         try:
             with open(filepath, "r") as json_file:
                 data = json.load(json_file)
         except:
-            print('{} not found. Did you forget --merge?'.format(filepath))
+            print("{} not found. Did you forget --merge?".format(filepath))
             exit(2)
 
-    compile_db = "{}/{}".format(remove_trailing(
-        args.dir, "/"), args.output)
+    compile_db = "{}/{}".format(remove_trailing(args.dir, "/"), args.output)
 
     if args.add_flags:
         data = add_flags(data, args.add_flags)
 
     if args.remove_files:
-        data = remove_files(data, [x.strip()
-                                   for x in args.remove_files.split(",")])
+        data = remove_files(
+            data, [args.path_prefix + x.strip() for x in args.remove_files.split(",")]
+        )
+
+    if args.include_files:
+        data = include_files(
+            data, [args.path_prefix + x.strip() for x in args.include_files.split(",")]
+        )
 
     if args.filter_files:
         data = filter_files(data, args.filter_files)
@@ -324,6 +387,9 @@ def main():
     if not args.allow_duplicates:
         data = [dict(t) for t in {tuple(d.items()) for d in data}]
 
+    if args.absolute_include_paths:
+        data = absolute_include_paths(data)
+
     overwrote = os.path.isfile(compile_db)
 
     if len(data) > 0:
@@ -331,7 +397,11 @@ def main():
             json.dump(data, json_file, indent=4, sort_keys=False)
     else:
         print("no commands found.")
-        exit(1)
+        if args.force_write:
+            with open(str(compile_db), "w") as json_file:
+                json.dump(data, json_file, indent=4, sort_keys=False)
+        else:
+            exit(1)
 
     if not args.quiet:
         end = time.time()
@@ -339,9 +409,9 @@ def main():
         print(
             "{} {} with {} command(s) in {}s.".format(
                 compile_db,
-                'updated' if overwrote else 'created',
+                "updated" if overwrote else "created",
                 len(data),
-                round(end - start, 4)
+                round(end - start, 4),
             )
         )
 
